@@ -1,3 +1,14 @@
+from typing import Any, Callable, Container, Sequence, Union
+
+import numpy as np
+
+import torch
+from torch.utils.data import SequentialSampler
+
+from catalyst.callbacks import IMetricCallback
+from catalyst.contrib.tools import SummaryWriter
+from catalyst.core import IRunner
+
 from dataclasses import dataclass, field
 from functools import partial
 from pathlib import Path
@@ -5,8 +16,9 @@ from typing import Callable, Dict, Optional, Tuple, List, Mapping
 
 import numpy as np
 import torch
+from catalyst.callbacks import BatchMetricCallback
 from catalyst.contrib.tools.tensorboard import SummaryWriter
-from catalyst.core import Callback, CallbackOrder, State, IRunner, BatchMetricCallback
+from catalyst.core import Callback, CallbackOrder, State, IRunner
 from torch import nn
 from torch.nn import DataParallel
 from torch.utils.data import Dataset, SequentialSampler
@@ -42,55 +54,43 @@ def publish_all(prefix: str,
     :param task_name: The name of the task, to be included in the name
     """
     if mapping_key in mapping:
-        publishers = mapping[mapping_key]
+        publishers: List[ITensorboardPublisher] = mapping[mapping_key]
         for publisher in publishers:
-            publisher(writer, idx, sample, prefix, task_name, key)
+            X, y = sample
+            tag = f'{prefix}_{task_name}'
+            publisher(writer, tag, X[key], idx)
 
 
-def img(writer: SummaryWriter, idx: int, sample: Tuple[Dict, Dict], prefix: str, task_name: str, key: str):
-    """
-    Publishes image interpretation data.
+class ITensorboardPublisher:
 
-    :param writer: A :class:`SummaryWriter` that logs to the tensorboard.
+    def __call__(self, writer: SummaryWriter, tag: str, sample: Any, idx: int):
+        """
+        Publishes interpretation data.
 
-    :param idx: The index of the dataset sample.
+        :param writer: A :class:`SummaryWriter` that logs to the tensorboard.
 
-    :param sample: A tuple of X, y dictionaries.
+        :param tag: The prefix, this is typically either "best_key" or "worst_key".
 
-    :param prefix: The prefix, this is typically either "best" or "worst".
+        :param sample: The sample that is going to be published to the Tensorboard
 
-    :param task_name: The name of the task, to be included in the name.
+        :param idx: The index of the dataset sample.
 
-    :param key: The key in X that is being published.
-    """
-    X, y = sample
-    writer.add_image(f'{prefix}_{task_name}_images', X[key], global_step=idx)
+        """
+        raise NotImplementedError()
 
 
-def text(writer: SummaryWriter, idx: int, sample: Tuple, prefix: str, task_name: str, key: str):
-    """
-    Publishes text interpretation data.
+def img_publisher(writer: SummaryWriter, tag: str, sample: Any, idx: int):
+    writer.add_image(f'{tag}_images', sample, global_step=idx)
 
-     :param writer: A :class:`SummaryWriter` that logs to the tensorboard.
 
-    :param idx: The index of the dataset sample.
-
-    :param sample: A tuple of X, y dictionaries.
-
-    :param prefix: The prefix, this is typically either "best" or "worst".
-
-    :param task_name: The name of the task, to be included in the name.
-
-    :param key: The key in X that is being published.
-    """
-    X, y = sample
-    writer.add_text(f'{prefix}_{task_name}_text', X[key], global_step=idx)
+def text_publisher(writer: SummaryWriter, tag: str, sample: Any, idx: int):
+    writer.add_text(f'{tag}_text', sample, global_step=idx)
 
 
 def default_tensorboard_type_mapping():
     return {
-        'img': [img],
-        'text': [text]
+        'img': [img_publisher],
+        'text': [text_publisher]
     }
 
 
@@ -100,14 +100,14 @@ class TensorboardConverter:
     A dataclass which holds mappings from column names to Tensorboard publishers and from column types to Tensorboard
     publishers. Also, it stores which column is of what type, to be able to log any column name to the Tensorboard.
     """
-    col_mapping: Dict[str, List[Callable]] = field(default_factory=lambda: {})
+    col_mapping: Dict[str, List[ITensorboardPublisher]] = field(default_factory=lambda: {})
     """
     Stores a `dict` from column names to a list of publishers. A publisher is just a callable which will be called 
     with this signature: 
     :code:`publisher(writer: SummaryWriter, idx: int, sample: Tuple, prefix: str, task_name: str, key: str)`. Example 
     publisher functions are :meth:`dnn_cool.catalyst_utils.img` and :meth:`dnn_cool.catalyst_utils.text`.
     """
-    type_mapping: Dict[str, List[Callable]] = field(default_factory=lambda: {})
+    type_mapping: Dict[str, List[ITensorboardPublisher]] = field(default_factory=lambda: {})
     """
     Stores a `dict` from column types to a list of publishers. A publisher is just a callable which will be called 
     with this signature: 
@@ -142,7 +142,14 @@ class TensorboardConverter:
             publish_all(prefix, idx, sample, key, key, writer, self.col_mapping, task_name)
         for key in X:
             if key in self.col_to_type_mapping:
-                publish_all(prefix, idx, sample, self.col_to_type_mapping[key], key, writer, self.type_mapping, task_name)
+                publish_all(prefix,
+                            idx,
+                            sample,
+                            self.col_to_type_mapping[key],
+                            key,
+                            writer,
+                            self.type_mapping,
+                            task_name)
 
 
 @dataclass
@@ -215,6 +222,7 @@ class InterpretationCallback(Callback):
     """
     This callback publishes best and worst images per task, according to the configuration supplied via the constructor.
     """
+
     def __init__(self, flow: TaskFlow,
                  tensorboard_converters: Optional[TensorboardConverters] = None,
                  loaders_to_skip=()):
@@ -460,3 +468,130 @@ def compute_device_metrics(reduced, any_tensor, metrics, outputs, targets):
                 value = value.unsqueeze(0)
             reduced[f'_device|{full_name}'] = value
         reduced[f'_device|{path}|_n'] = outputs[f'precondition|{metric.prefix}{metric.task_name}'].sum()
+
+
+class GuidedGradCamPublisher:
+
+    def __init__(self, model, layer, forward_pass_preprocess):
+        from captum.attr import GuidedGradCam
+        self.model = model
+        self.grad_cam = GuidedGradCam(model=model, layer=layer)
+        self.forward_pass_preprocess = forward_pass_preprocess
+
+    def __call__(self, writer: SummaryWriter, tag, sample, idx):
+        if not isinstance(sample, torch.Tensor):
+            sample = torch.tensor(sample)
+        if sample.shape[-1] == 3:  # H, W, C
+            sample = sample.permute(2, 0, 1)
+        X = self.forward_pass_preprocess(sample).unsqueeze(0)
+        device = next(self.model.parameters()).device
+        X = X.to(device)
+        logits = self.model(X)
+        res = self.grad_cam.attribute(X, logits.argmax())
+        res = res.squeeze().detach().cpu().numpy()
+        res = (res - res.min())
+        res /= res.max()
+        res = (res * 255).astype(np.uint8)
+        writer.add_image(f"{tag}_gradcam", res, global_step=idx)
+
+
+class SingleLossInterpretationCallback(IMetricCallback):
+    def __init__(
+            self,
+            criterion,
+            loaders_to_skip: Container[str] = (),
+            prefix: str = "",
+            input_key: str = "targets",
+            output_key: str = "logits",
+            idx_key=None,
+            top_k=10,
+            tensorboard_sequence: Sequence = None,
+            tensorboard_publishers: Sequence[ITensorboardPublisher] = (),
+            **loss_kwargs,
+    ):
+        super().__init__(prefix, input_key, output_key, **loss_kwargs)
+        self.metric = criterion
+        self.interpretations = {}
+        self.top_k = top_k
+        self.loggers = {}
+        self.tensorboard_sequence = tensorboard_sequence
+        self.tensorboard_publishers = tensorboard_publishers
+        self._loaders_to_skip = loaders_to_skip
+        self._idx_key = idx_key
+
+    def _should_interpret_loader(self, runner: IRunner):
+        if runner.loader_name in self._loaders_to_skip:
+            return False
+        if isinstance(runner.loaders[runner.loader_name].sampler, SequentialSampler):
+            return True
+
+        """
+        If the sampler is not sequential, we cannot recover the original index of the sample,
+        unless the user has provided `idx_key`.
+        See: https://github.com/catalyst-team/catalyst/issues/950#issuecomment-703220633
+        """
+        return self._idx_key is not None
+
+    def on_loader_start(self, runner: IRunner):
+        if not self._should_interpret_loader(runner):
+            return
+        if runner.loader_name not in self.loggers:
+            logdir = runner.logdir / f"{runner.loader_name}_log"
+            self.loggers[runner.loader_name] = SummaryWriter(str(logdir))
+        if runner.loader_name not in self.interpretations:
+            self.interpretations[runner.loader_name] = {
+                "loss": [],
+                "indices": [],
+            }
+
+    def on_loader_end(self, runner: IRunner):
+        if not self._should_interpret_loader(runner):
+            return
+
+        self.interpretations[runner.loader_name] = {
+            key: np.concatenate(value, axis=0)
+            for key, value in self.interpretations[runner.loader_name].items()
+        }
+
+        out_file = runner.logdir / f"{runner.loader_name}_interpretations.pkl"
+        torch.save(self.interpretations[runner.loader_name], out_file)
+
+        loss_sorter = self.interpretations[runner.loader_name]["loss"].argsort()
+        indices_sorted = self.interpretations[runner.loader_name]["indices"][loss_sorter]
+        indices = {
+            "best": indices_sorted[: self.top_k],
+            "worst": indices_sorted[-self.top_k:][::-1],
+        }
+
+        writer: SummaryWriter = self.loggers[runner.loader_name]
+        for type_prefix in ["best", "worst"]:
+            for idx in indices[type_prefix]:
+                tag = f"{self.prefix}{type_prefix}"
+                for tensorboard_publisher in self.tensorboard_publishers:
+                    sample = self.tensorboard_sequence[idx]
+                    tensorboard_publisher(writer, tag, sample, idx)
+
+    def on_batch_end(self, runner: IRunner):
+        if not self._should_interpret_loader(runner):
+            return
+        if self.metric is None:
+            return
+        loss_items: torch.Tensor = self._compute_metric_value(runner.output, runner.input)
+        if len(loss_items.shape) > 1:
+            dims = tuple(range(1, len(loss_items.shape)))
+            loss_items = loss_items.mean(dim=dims)
+
+        if self._idx_key is None:
+            bs = len(loss_items)
+            indices_so_far = self.interpretations[runner.loader_name]["indices"]
+            start_idx = (0 if len(indices_so_far) == 0 else (indices_so_far[-1][-1] + 1))
+            indices = np.arange(start_idx, start_idx + bs)
+        else:
+            indices = runner.input[self._idx_key].detach().cpu().numpy()
+
+        self.interpretations[runner.loader_name]["loss"].append(loss_items.detach().cpu().numpy())
+        self.interpretations[runner.loader_name]["indices"].append(indices)
+
+    @property
+    def metric_fn(self):
+        return self.metric
