@@ -56,7 +56,7 @@ class ITask:
     def get_loss(self):
         raise NotImplementedError()
 
-    def get_per_sample_loss(self):
+    def get_per_sample_loss(self, prefix='', ctx=None):
         raise NotImplementedError()
 
     def torch(self):
@@ -115,7 +115,7 @@ class Task(ITask):
     def get_loss(self):
         return self.loss
 
-    def get_per_sample_loss(self):
+    def get_per_sample_loss(self, prefix='', ctx=None):
         return self.per_sample_loss
 
     def torch(self):
@@ -232,12 +232,15 @@ class TaskFlow(ITask):
             self.tasks[task.get_name()] = task
         if flow_func is not None:
             self._flow_func = flow_func
+        self.ctx = {}
 
     def get_loss(self):
-        return TaskFlowLoss(self)
+        return TaskFlowLoss(self, ctx=self.ctx)
 
-    def get_per_sample_loss(self):
-        return TaskFlowLossPerSample(self)
+    def get_per_sample_loss(self, prefix='', ctx=None):
+        if ctx is None:
+            ctx = self.ctx
+        return TaskFlowLossPerSample(self, prefix=prefix, ctx=ctx)
 
     def torch(self):
         return TaskFlowModule(self)
