@@ -1,6 +1,8 @@
 import numpy as np
+import torch
 from sklearn.model_selection import train_test_split
 from torch.utils.data import Dataset, Subset
+from torchvision import transforms
 
 
 def any_value(outputs):
@@ -36,10 +38,10 @@ class TransformedSubset(Dataset):
         dataset (Dataset): The whole Dataset
         indices (sequence): Indices in the whole set selected for subset
     """
-    def __init__(self, dataset, indices, transforms=None):
+    def __init__(self, dataset, indices, sample_transforms=None):
         self.dataset = dataset
         self.indices = indices
-        self.transforms = transforms
+        self.transforms = sample_transforms
 
     def __getitem__(self, idx):
         r = self.dataset[self.indices[idx]]
@@ -65,3 +67,39 @@ def to_broadcastable_smaller(tensor1, tensor2):
     old_axes = tensor1.shape
     new_axes = [1] * (len(tensor2.shape) - len(tensor1.shape))
     return tensor1.view(*old_axes, *new_axes), tensor2
+
+
+def check_if_should_divide(sample):
+    if sample.dtype == np.uint8:
+        return True
+    if sample.dtype == torch.uint8:
+        return True
+    if sample.max() > 1.:
+        return True
+    return False
+
+
+def check_if_should_transpose(sample):
+    return sample.shape[-1] == 3
+
+
+class ImageNetNormalizer:
+
+    def __init__(self, should_divide_by_255=None, should_transpose=None):
+        self.should_divide_by_255 = should_divide_by_255
+        self.should_transpose = should_transpose
+        self.normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
+                                              std=[0.229, 0.224, 0.225])
+
+    def __call__(self, sample):
+        if self.should_divide_by_255 is None:
+            self.should_divide_by_255 = check_if_should_divide(sample)
+        if self.should_transpose is None:
+            self.should_transpose = check_if_should_transpose(sample)
+        if self.should_divide_by_255:
+            sample = sample / 255.
+        X = torch.tensor(sample).float()
+        if self.should_transpose:
+            X = X.permute(2, 0, 1)
+        X = self.normalize(X)
+        return X
